"""Vertical reduction operators."""

# Third-party
import numpy as np
import xarray as xr

# First-party
from meteodatalab.operators.destagger import destagger


def minmax_k(field, operator, mode, height, h_bounds, hsurf=None):
    """Find the extremum of a field given on k levels on some height interval.

    Parameters
    ----------
    field : xarray.DataArray
        field to reduce, defined either on model layer mid surfaces
        (typeOfLevel="generalVerticalLayer") or on model layer surfaces
        (typeOfLevel="generalVertical")
    height : xarray.DataArray
        height field defined on the same typeOfLevel as the field to reduce
    operator : str
        reduction operator, possible values are
        "maximum", "minimum"
    mode : str
        definition mode for height interval boundaries, possible values are
        "h2h": height agl to height agl
        "h2z": height agl to height amsl
        "z2h": height amsl to height agl
        "z2z": height amsl to height amsl
    h_bounds : list of xarray.DataArray of length 2
        height interval bounds (surface level field or single
        level of a multi-level field)
    hsurf : Optional(xarray.DataArray)
        earth surface height in m amsl
        required if mode is one of {"h2h", "h2z", "z2h"}

    Returns
    -------
    rfield : xarray.DataArray
        reduced field

    """
    # Note on multilevel results
    # z2z: h_bounds elements refer to single level of a height field
    # z2h: h_bounds elements refer to single level of a height field
    # h2h: h_bounds elements refer to single level of a height field
    # h2z: h_bounds elements refer to single level of a height field
    # TODO: add some checks on level types of input fields, and on values
    # and order of h_bounds

    # Parameters
    # ... supported reduction operators
    reduction_operators = ("maximum", "minimum")
    # ... supported height interval modes
    height_interval_modes = ("h2h", "h2z", "z2h", "z2z")

    # Check arguments
    # ... operator
    if operator not in reduction_operators:
        raise RuntimeError("minmax_k: unsupported operator ", operator)
    # ... mode
    if mode not in height_interval_modes:
        raise RuntimeError("minmax_k: unsupported mode ", mode)
    # ... hsurf
    if mode in ["h2z", "z2h", "h2h"] and hsurf is None:
        raise RuntimeError(
            "minmax_k: hsurf is required when using operator ",
            operator,
            "with mode ",
            mode,
        )

    # Height bounds are always converted to heights amsl; lower_bound_type
    # and upper_bound_type can be used to code typeOfFirstFixedSurface
    # and typeOfSecondFixedSurface in GRIB2, and to set the bounds attribute for
    # the vertical coordinates in NetCDF
    h_bottom, h_top = h_bounds
    if mode in ["h2z", "h2h"]:
        # raise error as long as no unit test is available
        raise NotImplementedError(
            "minmax_k: unit test not yet implemented for mode ", mode
        )
        # ... convert lower bound to height amsl: h_bottom += hsurf
    if mode in ["z2h", "h2h"]:
        # raise error as long as no unit test is available
        raise NotImplementedError(
            "minmax_k: unit test not yet implemented for mode ", mode
        )
        # ... convert upper bound to height amsl: h_top += hsurf

    # Find height interval including the interval [h_bottom, h_top]
    # ... maximum/minimum: extremum over the field values at all model
    #  levels included in the height interval, and at the interval boundaries
    #     after linear interpolation wrt height; f and auxiliary height fields
    #     must either both be defined on full levels or half levels
    if field.vcoord_type != height.vcoord_type or field.origin_z != height.origin_z:
        raise RuntimeError(
            "minmax_k: height is not defined for the same level type as field."
        )
    field_in_h_bounds = field.where((height >= h_bottom) & (height <= h_top)).dropna(
        "z"
    )
    heightkp1 = height.shift({"z": -1})
    fieldkp1 = field.shift({"z": -1})
    gradf = (field - fieldkp1) / (height - heightkp1)
    gradfkm1 = gradf.shift({"z": 1})
    field_extrapolated_to_h_top = xr.where(
        (height > h_top) & (heightkp1 < h_top),
        field + gradf * (height - h_top),
        np.nan,
    ).dropna("z")
    field_extrapolated_to_h_bottom = xr.where(
        (heightkp1 < h_bottom) & (height > h_bottom),
        field + gradfkm1 * (height - h_bottom),
        np.nan,
    ).dropna("z")

    # ... compute the extremum
    if operator == "minimum":
        rfield = field_in_h_bounds.min(dim="z")
        if field_extrapolated_to_h_bottom.size > 0:
            rfield = xr.where(
                field_extrapolated_to_h_bottom < rfield,
                field_extrapolated_to_h_bottom,
                rfield,
            )
        if field_extrapolated_to_h_top.size > 0:
            rfield = xr.where(
                field_extrapolated_to_h_top < rfield,
                field_extrapolated_to_h_top,
                rfield,
            )
    else:
        rfield = field_in_h_bounds.max(dim="z")
        if field_extrapolated_to_h_bottom.size > 0:
            rfield = xr.where(
                field_extrapolated_to_h_bottom > rfield,
                field_extrapolated_to_h_bottom,
                rfield,
            )
        if field_extrapolated_to_h_top.size > 0:
            rfield = xr.where(
                field_extrapolated_to_h_top > rfield,
                field_extrapolated_to_h_top,
                rfield,
            )

    return rfield


def integrate_k(field, operator, mode, height, h_bounds, hsurf=None):
    """Integrate a field given on k levels on some height interval.

    Parameters
    ----------
    field : xarray.DataArray
        field to reduce, defined either on model layer mid
        surfaces (typeOfLevel="generalVerticalLayer")
        or on model layer surfaces (typeOfLevel="generalVertical")
    height : xarray.DataArray
        height field on model layer surfaces (typeOfLevel="generalVertical")
    operator : str
        integral operator, possible values are
        "integral", "normed_integral"
    mode : str
        definition mode for height interval boundaries, possible values are
        "h2h": height agl to height agl
        "h2z": height agl to height amsl
        "z2h": height amsl to height agl
        "z2z": height amsl to height amsl
    h_bounds : list of xarray.DataArray of length 2 height interval bounds
        (surface level field or single level of a multi-level field)
    hsurf : Optional(xarray.DataArray)
        earth surface height in m amsl
        required if mode is one of {"h2h", "h2z", "z2h"}

    Returns
    -------
    rfield : xarray.DataArray
        reduced field

    """
    # Note on multilevel results
    # z2z: h_bounds elements refer to single level of a height field
    # z2h: h_bounds elements refer to single level of a height field
    # h2h: h_bounds elements refer to single level of a height field
    # h2z: h_bounds elements refer to single level of a height field
    # TODO: add some checks on level types of input fields, and on values and
    # order of h_bounds

    # Parameters
    # ... supported reduction operators
    integral_operators = ("integral", "normed_integral")
    # ... supported height interval modes
    height_interval_modes = ("h2h", "h2z", "z2h", "z2z")

    # Check arguments
    # ... operator
    if operator not in integral_operators:
        raise RuntimeError("integrate_k: unsupported operator ", operator)
    # ... mode
    if mode not in height_interval_modes:
        raise RuntimeError("integrate_k: unsupported mode ", mode)
    # ... hsurf
    if mode in ["h2z", "z2h", "h2h"] and hsurf is None:
        raise RuntimeError(
            "integrate_k: hsurf is required when using operator ",
            operator,
            "with mode ",
            mode,
        )

    # Height bounds are always converted to heights a msl
    # TODO: additional variables lower_bound_type and upper_bound_type could be set
    # here for later usage to code
    # typeOfFirstFixedSurface and typeOfSecondFixedSurface in GRIB2, and to set the
    # bounds attribute for the vertical
    # coordinates in NetCDF
    h_bottom, h_top = h_bounds
    if mode in ["h2z", "h2h"]:
        # raise error as long as no unit test is available
        raise NotImplementedError(
            "integrate_k: unit test not yet implemented for mode ", mode
        )
        # ... convert lower bound to height amsl: h_bottom += hsurf
    if mode in ["z2h", "h2h"]:
        # raise error as long as no unit test is available
        raise NotImplementedError(
            "integrate_k: unit test not yet implemented for mode ", mode
        )
        # ... convert upper bound to height amsl: h_top += hsurf

    # Find height interval including the interval [h_bottom, h_top]
    # ... integral: approximated by midpoint rule, taking into account that h_bottom
    #         and h_top, respectively, may not coincide with a model layer interface;
    #
    #       if typeOfLevel(f)="generalVerticalLayer"
    #         f(kstart)[h_top - hhl(kstart+1)] +
    #         sum(k=kstart+1,kstop-1)f(k)[hhl(k)-hhl(k+1)] +
    #         f(kstop)[hhl(kstop) - h_bottom]
    #       if typeOfLevel(f)="generalVertical"
    #         0.5*[f(kstart+1) + f(kstart)][h_top - hhl(kstart+1)] +
    #         sum(k=kstart+1,kstop-1)0.5*[f(k+1)+f(k)][hhl(k)-hhl(k+1)] +
    #         0.5*[f(kstop)+f(kstop+1)][hhl(kstop) - h_bottom]
    #       kstart and kstop refer to all model midpoint surfaces included
    #       in the height interval.
    # ... normed_integral: integral / (h_top - h_bottom)
    if field.vcoord_type != "model_level":
        raise RuntimeError(
            "integrate_k: field must be defined for level type "
            "generalVertical or generalVerticalLayer"
        )
    if field.origin_z != 0.0:
        field_on_fl = destagger(field, "z")
    else:
        field_on_fl = field
    # ... prepare the height hfl of model mid layer surfaces (needed to select
    # the field values within [h_bottom, h_top])
    hfl = destagger(height, "z")
    # .. prepare dh = height(k) - height(k+1), defined on model mid layer surfaces
    hhlk = height[{"z": slice(-1)}]
    hhlkp1 = height.shift(z=-1)[{"z": slice(-1)}]
    dh = xr.where((hhlk > h_top) & (hhlkp1 < h_top), h_top - hhlkp1, hhlk - hhlkp1)
    dh = xr.where((hhlkp1 < h_bottom) & (hhlk > h_bottom), hhlk - h_bottom, dh)

    # ... find field and dh where hfl is in interval [h_bottom, h_top]
    # ... note that the dimension "generalVericalLayer" is lost of this condition is
    # nowhere satisfied
    field_in_h_bounds = field_on_fl.where((hfl >= h_bottom) & (hfl <= h_top)).dropna(
        dim="z",
        how="all",
    )
    dh_in_h_bounds = dh.where((hfl >= h_bottom) & (hfl <= h_top)).dropna(
        dim="z",
        how="all",
    )
    # ... compute integral by midpoint rule (apply fractional corrections for
    # the height intervals containing h_top and h_bottom)
    #     at grid points where field_in_h_bounds is not undefined for all entries
    #     along dimension "generalVerticalLayer"
    # NOTE: The vertical dimension is lost in the reduction operation; one could use
    # xr.DataArray.expand_dims to add a vertical dim
    #       of size 1 and assign a coordinate with associated attributes to it
    #       re-ordering of dimensions would, however, be unnecessary due to xarray's
    #       broadcasting by dimension name
    # TODO: assign coordinates and attributes to rfield
    if "z" in field_in_h_bounds.dims:
        rfield = (
            (field_in_h_bounds * dh_in_h_bounds)
            .sum(dim="z")
            .where(~field_in_h_bounds.isnull().all(dim="z"))
            # the line above reverts all nan columns to nan value instead of zero
        )
        if operator == "normed_integral":
            rfield /= h_top - h_bottom
    else:
        rfield = xr.full_like(field_in_h_bounds, fill_value=np.nan)

    return rfield



"""Algorithm for the curl operator."""

# Standard library
from typing import cast

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import physical_constants as pc
from ..metadata import is_staggered_horizontal
from . import diff
from .destagger import destagger
from .total_diff import TotalDiff


def curl(
    u: xr.DataArray,
    v: xr.DataArray,
    w: xr.DataArray,
    rlat: xr.DataArray,
    total_diff: TotalDiff,
) -> tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
    """Compute the curl of the velocity field."""
    r_earth_inv = 1 / pc.earth_radius
    acrlat = cast(xr.DataArray, 1 / (np.cos(rlat) * pc.earth_radius))
    tgrlat = cast(xr.DataArray, np.tan(rlat))

    # compute weighted derivatives for FD
    u_f = destagger(u, "x") if is_staggered_horizontal(u) else u
    v_f = destagger(v, "y") if is_staggered_horizontal(v) else v
    w_f = destagger(w, "z")

    du_dz = diff.dz(u_f)
    du_dphi = total_diff.d_dphi(diff.dy(u_f), du_dz)
    du_dzeta = total_diff.d_dzeta(du_dz)

    dv_dz = diff.dz(v_f)
    dv_dlam = total_diff.d_dlam(diff.dx(v_f), dv_dz)
    dv_dzeta = total_diff.d_dzeta(dv_dz)

    dw_dz = diff.dz_staggered(w)
    dw_dlam = total_diff.d_dlam(diff.dx(w_f), dw_dz)
    dw_dphi = total_diff.d_dphi(diff.dy(w_f), dw_dz)

    # compute curl
    curl1 = acrlat * (r_earth_inv * dw_dphi + dv_dzeta - r_earth_inv * v_f)
    curl2 = r_earth_inv * (-du_dzeta - acrlat * dw_dlam + r_earth_inv * u_f)
    curl3 = acrlat * dv_dlam + r_earth_inv * (-du_dphi + tgrlat * u_f)

    return curl1, curl2, curl3


def curl_alt(
    u: xr.DataArray,
    v: xr.DataArray,
    w: xr.DataArray,
    rlat: xr.DataArray,
    total_diff: TotalDiff,
) -> tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
    td = total_diff

    r_earth_inv = 1 / pc.earth_radius
    acrlat = 1 / (np.cos(rlat) * pc.earth_radius)
    tgrlat = np.tan(rlat)

    u_f = destagger(u, "x") if is_staggered_horizontal(u) else u
    v_f = destagger(v, "y") if is_staggered_horizontal(v) else v
    w_f = destagger(w, "z")

    du_dphi = np.gradient(u_f, td.dlat, axis=-2)
    du_dzeta = np.gradient(u_f, axis=-3)

    dv_dlam = np.gradient(v_f, td.dlon, axis=-1)
    dv_dzeta = np.gradient(v_f, axis=-3)

    dw_dlam = np.gradient(w_f, td.dlon, axis=-1)
    dw_dphi = np.gradient(w_f, td.dlat, axis=-2)
    dw_dzeta = np.diff(w, axis=-3)

    curl1 = acrlat * (
        r_earth_inv * (dw_dphi + td.dzeta_dphi * dw_dzeta)
        + td.sqrtg_r_s * dv_dzeta
        - r_earth_inv * v_f
    )
    curl2 = r_earth_inv * (
        -td.sqrtg_r_s * du_dzeta
        - acrlat * (dw_dlam + td.dzeta_dlam * dw_dzeta)
        + r_earth_inv * u_f
    )
    curl3 = (
        acrlat * (dv_dlam + td.dzeta_dlam * dv_dzeta)
        - r_earth_inv * (du_dphi + td.dzeta_dphi * du_dzeta)
        + r_earth_inv * tgrlat * u_f
    )

    return curl1, curl2, curl3



"""Relative humidity operators."""

# Standard library
from typing import Literal

# Third-party
import xarray as xr

# Local
from .. import metadata
from .atmo import pv_si, pv_sm, pv_sw, qv_pvp


def relhum(
    qv, t, p, clipping=True, phase: Literal["water", "ice", "water+ice"] = "water"
):
    """Calculate relative humidity.

    Parameters
    ----------
    qv : xarray.DataArray
        water vapor mixing ratio
    t : xarray.DataArray
        temperature in Kelvin
    p : xarray.DataArray
        pressure in Pa
    clipping : bool
        clips the relative humidity to [0,100] interval.
        Only upper bound is controlled by this parameter,
        since lower bound clipping is always performed.
    phase : Literal["water", "ice", "water+ic"]
        Customizes how relative humidity is computed.
        'water'        over water
        'ice'          over ice
        'water+ice'    over mixed phase

    Returns
    -------
    xarray.DataArray
        relative humidity field in %

    """
    max = 100 if clipping else None

    phase_conditions = {
        "water": {"func": pv_sw(t), "shortName": "RELHUM"},
        "ice": {"func": pv_si(t), "shortName": "RH_ICE"},
        "water+ice": {"func": pv_sm(t), "shortName": "RH_MIX_EC"},
    }

    if phase not in phase_conditions:
        raise ValueError("Invalid phase. Phase must be 'water', 'ice', or 'water+ice'.")

    result = (100.0 * qv / qv_pvp(phase_conditions[phase]["func"], p)).clip(0, max)

    return xr.DataArray(
        data=result,
        attrs=metadata.override(
            t.metadata, shortName=phase_conditions[phase]["shortName"]
        ),
    )



"""Data source helper class."""

# Standard library
import dataclasses as dc
import os
import typing
from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager, nullcontext
from contextvars import ContextVar
from functools import singledispatchmethod
from pathlib import Path

# Third-party
import earthkit.data as ekd  # type: ignore
import eccodes  # type: ignore
import eccodes_cosmo_resources

try:
    # Third-party
    from polytope.api import Client as PolytopeClient  # type: ignore
except ImportError:
    PolytopeClient = None

# Local
from . import config, mars

definitions = ContextVar("definitions", default="vendor")


@contextmanager
def cosmo_grib_defs():
    """Enable COSMO GRIB definitions."""
    if definitions.get() == "cosmo":
        yield
        return

    token = definitions.set("cosmo")
    restore = eccodes.codes_definition_path()
    paths = (
        eccodes_cosmo_resources.get_definitions_path(),
        Path(restore),
    )
    for path in paths:
        if not path.exists() and not str(path).startswith("/MEMFS"):
            raise RuntimeError(f"{path} does not exist")
    defs_path = ":".join(map(str, paths))
    eccodes.codes_set_definitions_path(defs_path)
    try:
        yield
    finally:
        definitions.reset(token)
        eccodes.codes_set_definitions_path(restore)


def grib_def_ctx(grib_def: str):
    if "ECCODES_DEFINITION_PATH" in os.environ or "GRIB_DEFINITION_PATH" in os.environ:
        return nullcontext()
    if grib_def == "cosmo":
        return cosmo_grib_defs()
    return nullcontext()


@dc.dataclass
class DataSource(ABC):
    request_template: dict[str, typing.Any] = dc.field(default_factory=dict)

    @singledispatchmethod
    def retrieve(
        self, request: dict[str, typing.Any] | str | tuple[str, str] | mars.Request
    ) -> Iterator:
        """Stream GRIB fields from files or FDB.

        Request for data from the source in the mars language.

        Simple strings are interpreted as `param` filters and pairs of strings
        are interpreted as `param` and `levtype` filters.

        Key value pairs from the `request_template` attribute are used as default
        values. Note that the default values in the mars request passed as an input
        will take precedence on the template values.

        Parameters
        ----------
        request : dict | str | tuple[str, str] | meteodatalab.mars.Request
            Request for data from the source in the mars language.

        Yields
        ------
        GribField
            GribField instances containing the requested data.

        """
        raise NotImplementedError(f"request of type {type(request)} not supported.")

    @retrieve.register
    def _(self, request: dict) -> Iterator:
        # The presence of the yield keyword makes this def a generator.
        # As a result, the context manager will remain active until the
        # exhaustion of the data source iterator.
        grib_def = config.get("data_scope", "cosmo")
        with grib_def_ctx(grib_def):
            yield from self._retrieve(request)

    @retrieve.register
    def _(self, request: mars.Request) -> Iterator:
        yield from self.retrieve(request.dump())

    @retrieve.register
    def _(self, request: str) -> Iterator:
        yield from self.retrieve({"param": request})

    @retrieve.register
    def _(self, request: tuple) -> Iterator:
        param, levtype = request
        yield from self.retrieve({"param": param, "levtype": levtype})

    @abstractmethod
    def _retrieve(self, request: dict):
        pass


@dc.dataclass
class FDBDataSource(DataSource):
    def _retrieve(self, request: dict):
        req_kwargs = self.request_template | request
        req = mars.Request(**req_kwargs)
        yield from ekd.from_source("fdb", req.to_fdb())


@dc.dataclass
class FileDataSource(DataSource):
    datafiles: list[str] | None = None

    def _retrieve(self, request: dict):
        req_kwargs = self.request_template | request
        _ = mars.Request(**req_kwargs)
        fs = ekd.from_source("file", self.datafiles)
        yield from fs.sel(req_kwargs)


@dc.dataclass
class PolytopeDataSource(DataSource):
    polytope_collection: str | None = None
    polytope_client: PolytopeClient = dc.field(init=False)

    def __post_init__(self):
        if PolytopeClient is None:
            raise ImportError(
                "Please install polytope-client to use the polytope data source"
            )
        self.polytope_client = PolytopeClient()

    def _retrieve(self, request: dict):
        req_kwargs = self.request_template | request
        req = mars.Request(**req_kwargs)
        pointers = self.polytope_client.retrieve(
            self.polytope_collection,
            req.to_polytope(),
            pointer=True,
            asynchronous=False,
        )
        urls = [p["location"] for p in pointers]
        yield from ekd.from_source("url", urls, stream=True)


@dc.dataclass
class URLDataSource(DataSource):
    urls: list[str] | None = None

    def _retrieve(self, request: dict):
        req_kwargs = self.request_template | request
        fs = ekd.from_source("url", self.urls)
        yield from fs.sel(**req_kwargs)



"""Finite difference stencils on xarray dataarrays."""

# Standard library
import dataclasses as dc

# Third-party
import numpy as np
import xarray as xr

# Local
from . import diff
from .destagger import destagger


@dc.dataclass
class TotalDiff:
    """Diff operators for terrain following grid."""

    dlon: float
    dlat: float
    sqrtg_r_s: xr.DataArray
    dzeta_dlam: xr.DataArray
    dzeta_dphi: xr.DataArray

    @classmethod
    def from_hhl(cls, hhl: xr.DataArray):
        deg2rad = np.pi / 180
        dlon = hhl.geography["iDirectionIncrementInDegrees"] * deg2rad
        dlat = hhl.geography["jDirectionIncrementInDegrees"] * deg2rad

        dh_dx = destagger(diff.dx(hhl), "z")  # order is important
        dh_dy = destagger(diff.dy(hhl), "z")  # diff then destagger
        dh_dz = diff.dz_staggered(hhl)

        sqrtg_r_s = -1 / dh_dz
        dzeta_dlam = sqrtg_r_s / dlon * dh_dx
        dzeta_dphi = sqrtg_r_s / dlat * dh_dy

        return cls(dlon, dlat, sqrtg_r_s, dzeta_dlam, dzeta_dphi)

    def d_dlam(self, df_dx: xr.DataArray, df_dz: xr.DataArray) -> xr.DataArray:
        """Compute the derivative along the lambda axis."""
        return df_dx / self.dlon + df_dz * self.dzeta_dlam

    def d_dphi(self, df_dy: xr.DataArray, df_dz: xr.DataArray) -> xr.DataArray:
        """Compute the derivative along the phi axis."""
        return df_dy / self.dlat + df_dz * self.dzeta_dphi

    def d_dzeta(self, df_dz: xr.DataArray) -> xr.DataArray:
        """Compute the derivative along the zeta axis."""
        return df_dz * self.sqrtg_r_s



"""Command line interface of meteodata-lab."""

# Standard library
from importlib.resources import files
from pathlib import Path

# Third-party
import click
import yaml

# Local
from . import __version__, grib_decoder
from .metadata import is_staggered_horizontal
from .operators import destagger, gis, regrid


def print_version(ctx, _, value: bool) -> None:
    """Print the version number and exit."""
    if value:
        click.echo(__version__)
        ctx.exit(0)


@click.group()
@click.option(
    "--version",
    "-V",
    help="Print version and exit.",
    is_flag=True,
    expose_value=False,
    callback=print_version,
)
def main() -> None:
    """Console script for test_cli_project."""
    print("CLI for meteodatalab")


RESAMPLING = {
    "nearest": regrid.Resampling.nearest,
    "bilinear": regrid.Resampling.bilinear,
    "cubic": regrid.Resampling.cubic,
}


def _load_mapping():
    mapping_path = files("meteodatalab.data").joinpath("field_mappings.yml")
    return yaml.safe_load(mapping_path.open())


def handle_vector_fields(ds):
    mapping = _load_mapping()
    names = set(ds)
    pairs = []
    while names:
        name = names.pop()
        item = mapping[name]["cosmo"]
        if u := item.get("uComponent"):
            if u not in names:
                msg = f"The u-component {u} must be part of PARAMS"
                raise click.UsageError(msg)
            names.remove(u)
            pairs.append((u, name))
        elif v := item.get("vComponent"):
            if v not in names:
                msg = f"The v-component {v} must be part of PARAMS"
                raise click.UsageError(msg)
            names.remove(v)
            pairs.append((name, v))

    for u_name, v_name in pairs:
        click.echo(f"Rotating vector field components {u_name}, {v_name} to geolatlon")
        u, v = ds[u_name], ds[v_name]
        if is_staggered_horizontal(u):
            u = destagger.destagger(u, "x")
        if is_staggered_horizontal(v):
            v = destagger.destagger(v, "y")
        if u.vref == "native" and v.vref == "native":
            u_g, v_g = gis.vref_rot2geolatlon(u, v)
            ds[u_name] = u_g
            ds[v_name] = v_g
        elif u.vref == "geo" and v.vref == "geo":
            continue
        else:
            msg = "Differing vector references."
            raise RuntimeError(msg)


@main.command("regrid")
@click.option(
    "--crs",
    type=click.Choice(["geolatlon"]),
    default="geolatlon",
    help="Coordinate reference system",
)
@click.option(
    "--resampling",
    type=click.Choice(list(RESAMPLING.keys())),
    default="nearest",
    help="Resampling method",
)
@click.option(
    "--ref-param",
    default="HHL",
    help="Reference parameter with respect to staggering, default: HHL",
)
@click.argument(
    "infile",
    nargs=-1,
    type=click.Path(exists=True, path_type=Path),
)
@click.argument(
    "outfile",
    type=click.Path(writable=True, path_type=Path),
)
@click.argument("params")
def regrid_cmd(
    crs: str,
    resampling: str,
    ref_param: str,
    infile: tuple[Path, ...],
    outfile: Path,
    params: str,
):
    """Regrid the given PARAMS found in INFILE and write to OUTFILE.

    PARAMS is a comma separated list of short names following the cosmo convention.
    """
    resampling_arg = RESAMPLING[resampling]
    crs_str = regrid.CRS_ALIASES.get(crs, crs)

    if not infile:
        raise click.UsageError("No input files")

    if outfile.exists():
        click.confirm(f"OUTFILE {outfile} exists. Overwrite?")

    reader = grib_decoder.GribReader.from_files(list(infile), ref_param=ref_param)
    ds = reader.load_fieldnames(params.split(","))

    handle_vector_fields(ds)

    with outfile.open("wb") as fout:
        for name, field in ds.items():
            src = regrid.RegularGrid.from_field(field)
            dst = src.to_crs(crs_str)

            click.echo(f"Regridding field {name} to {dst}")
            field_out = regrid.regrid(field, dst, resampling_arg)

            click.echo(f"Writing GRIB fields to {outfile}")
            grib_decoder.save(field_out, fout)
